package de.proteinms.xtandemparser.Mutater;


import java.util.Vector;

/**
 * Created by IntelliJ IDEA.
 * User: niklaas
 * Date: 04-Aug-2009
 * Time: 16:37:06
 */
public class MutationFinder {
    /**
     * The peptide to mutate
     */
    private Element[] iPeptide;

    /**
     * The resulting mutated peptides
     */
    private Vector<MutatedPeptide> newSeqVector = new Vector<MutatedPeptide> ();

    /**
     * The mass limit
     */
    private double iMassLimit;
    private double iPeptideMass;
    private String iPeptideSequence = "";

    public MutationFinder(Element[] aPeptide, double aMassLimit){
        this.iPeptide = aPeptide;
        for(int i = 0; i< iPeptide.length; i ++){
            if(iPeptide[i] == null){
                System.out.println("Error");
            } else {
                iPeptideSequence = iPeptideSequence + iPeptide[i].getElement();
            }
        }
        this.iMassLimit = aMassLimit;
        this.iPeptideMass = getMass(aPeptide);
    }

    public double getMass(Element[] lElements){
        double lResult = 0.0;
        for(int i = 0; i<lElements.length; i ++){
            if(lElements[i] != null){
                lResult = lResult + lElements[i].getMw();
            }
        }
        return lResult;
    }

   public double getMass(Vector<Element> lElements){
        double lResult = 0.0;
        for(int i = 0; i<lElements.size(); i ++){
            if(lElements.get(i) != null){
                lResult = lResult + lElements.get(i).getMw();
            }
        }
        return lResult;
    }

    /**
     * This method will find insertions
     * @param aCoupleToFind
     * @param aCoupleToReplace
     * @param aFindType
     * @param aReplaceType
     */
    public void findInsertion(Vector<Couple> aCoupleToFind, Vector<Couple> aCoupleToReplace, int aFindType, int aReplaceType){
        //create a couple array
        Couple[] aCoupleMap = new Couple[aCoupleToFind.size()];
        aCoupleToFind.toArray(aCoupleMap);

        //check every couple if we can create insertion in it
        for(int j = 0 ; j<aCoupleMap.length ; j++){
            //get the couple
            Couple couple = aCoupleMap[j];
            Element ele1 = couple.getElement1();
            double lMass = couple.getMw();


            //create the boolean that indicates if we found element1
            boolean lFoundEle1 = false;
            //the position of element 1
            int lPosition1 = 0;

            for(int k = 0; k<iPeptide.length; k ++){
                if(ele1.getElement().equalsIgnoreCase(iPeptide[k].getElement())){
                    if(k != 0 && k!= iPeptide.length -1){
                        //check if the found position can match the type
                        lFoundEle1 = true;
                        lPosition1 = k;
                        //we found the element
                        //check if we can find a couple in a specific mass range for an insertion
                        Vector<Couple> lReplaceCouples = getCouplesInRange(aCoupleToReplace,lMass, iMassLimit,couple);
                        for(int i = 0 ; i<lReplaceCouples.size() ; i++){
                            Couple coupleFound = lReplaceCouples.get(i);
                            Vector<Element> seque = new Vector<Element>();
                            boolean lUse = false;
                            if(aReplaceType == 22 && aFindType == 2 ){
                                lUse = true;
                            }
                            if(aReplaceType == 12 && aFindType == 1 ){
                                lUse = true;
                            }
                            if(aReplaceType == 23 && aFindType == 3 ){
                                lUse = true;
                            }
                            if(lUse){
                                String newSequence ="";
                                for(int l = 0; l<iPeptide.length ; l++ ){
                                    if(l == lPosition1){
                                        seque.add(coupleFound.getElement1());
                                        seque.add(coupleFound.getElement2());
                                        newSequence =  newSequence + coupleFound.getElement1().getElement() + coupleFound.getElement2().getElement();
                                    } else {
                                        seque.add(iPeptide[l]);
                                        newSequence =  newSequence + iPeptide[l].getElement();
                                    }
                                }

                                Element[] lResult = new Element[seque.size()];
                                seque.toArray(lResult);
                                newSeqVector.add(new MutatedPeptide(iPeptideSequence, newSequence, "i", iPeptide[lPosition1].getElement() + " |"  + lPosition1 + "| " + coupleFound.getElement1().getElement() + coupleFound.getElement2().getElement() ,   "i" , iPeptideMass,getMass(seque), lResult));
                            }
                        }
                    }
                }
            }
        }
    }


    public Vector<Couple> getCouplesInRange(Vector<Couple> lCouples, double lMass, double lRange, Couple lOriginalCouple){
        int lMiddle = 0;
        int lMiddleLast = -1;
        int lMin = 0;
        int lMax = lCouples.size() - 1;
        boolean lCoupleFound = false;
        Couple lCouple = null;
        boolean lContinue = true;
        Vector<Couple> lResults = new Vector<Couple>();


        while(lContinue){
            lMiddle = (lMin + lMax) / 2;
            if(lMiddle != lMiddleLast){
                lCouple = lCouples.get(lMiddle);
                double lCoupleMass = lCouple.getMw();
                if(lMass - lRange <= lCoupleMass && lMass + lRange >= lCoupleMass){
                    lCoupleFound = true;
                    lContinue = false;
                } else {
                    if(lCoupleMass > lMass){
                        lMax = lMiddle;
                    } else if(lCoupleMass < lMass){
                        lMin = lMiddle;
                    }
                }
                if(lMin >= lMax){
                    lContinue = false;
                }
                lMiddleLast = lMiddle;
            } else {
                if(lMin == 0){
                    //still zero, we should check the first couple
                    lCouple = lCouples.get(0);
                    double lCoupleMass = lCouple.getMw();
                    if(lMass - lRange <= lCoupleMass && lMass + lRange >= lCoupleMass){
                        lCoupleFound = true;
                    }
                } else if(lMax == lCouples.size() - 1){
                    //still max, we should check the last couple
                    lCouple = lCouples.get(lMax);
                    double lCoupleMass = lCouple.getMw();
                    if(lMass - lRange <= lCoupleMass && lMass + lRange >= lCoupleMass){
                        lCoupleFound = true;
                    }
                }
                //we are at the end of the search
                lContinue = false;
            }
        }


        if(lCoupleFound){
            lResults.add(lCouple);
            boolean lAdd = true;
            int lMiddleUp = lMiddle;
            int lMiddleDown = lMiddle;
            //try to add couples larger then the current selection
            while(lAdd){
                lMiddleUp = lMiddleUp + 1;
                if(lMiddleUp <lCouples.size()){
                    Couple lAddCouple = lCouples.get(lMiddleUp);
                    double lCoupleMass = lAddCouple.getMw();
                    if(lMass - lRange <= lCoupleMass && lMass + lRange >= lCoupleMass){
                        lResults.add(lAddCouple);
                    } else {
                        lAdd = false;
                    }
                } else {
                    lAdd = false;
                }
            }
            //try to add couples smaller then the current selection
            lAdd = true;
            while(lAdd){
                lMiddleDown = lMiddleDown - 1;
                if(lMiddleDown >= 0){
                    Couple lAddCouple = lCouples.get(lMiddleDown);
                    double lCoupleMass = lAddCouple.getMw();
                    if(lMass - lRange <= lCoupleMass && lMass + lRange >= lCoupleMass){
                        lResults.add(lAddCouple);
                    } else {
                        lAdd = false;
                    }
                } else {
                    lAdd = false;
                }
            }

        }
        lResults.remove(lOriginalCouple);

        return lResults;
    }

    /**
     * This method will find deletions
     * @param aCoupleToFind
     * @param aCoupleToReplace
     * @param aFindType
     * @param aReplaceType
     */
    public void findDeletion(Vector<Couple> aCoupleToFind, Vector<Couple> aCoupleToReplace, int aFindType, int aReplaceType){
        //create a couple array
        Couple[] aCoupleMap = new Couple[aCoupleToFind.size()];
        aCoupleToFind.toArray(aCoupleMap);

        //check every couple if we can create deletions in it
        for(int j = 0 ; j<aCoupleMap.length ; j++){
            //get the couple
            Couple couple = aCoupleMap[j];
            Element ele1 = couple.getElement1();
            Element ele2 = couple.getElement2();
            double lMass = couple.getMw();


            //create the boolean that indicates if we found element1
            boolean lFoundEle1 = false;
            //create the boolean that indicates if we found element2
            boolean lFoundEle2 = false;
            //the position of element 1
            int lPosition1 = 0;
            //the position of element 2
            int lPosition2 = 0;

            for(int k = 0; k<iPeptide.length; k ++){
                if(ele1.getElement().equalsIgnoreCase(iPeptide[k].getElement())){
                    //check if the found position is correct with the type
                    if(aFindType == 12 || aFindType == 13){
                        if(k == 0){
                            lFoundEle1 = true;
                            lPosition1 = k;
                        }
                    }
                    if(aFindType == 22  || aFindType == 23){
                        if(k != 0 && k!= iPeptide.length -1){
                            lFoundEle1 = true;
                            lPosition1 = k;
                        }
                    }
                }
                if(lFoundEle1){
                    //we found the first element, try to find the second one
                    for(int l = 0; l<iPeptide.length; l ++){
                        if(ele2.getElement().equalsIgnoreCase(iPeptide[l].getElement())){
                            //check if the found position is correct with the type
                            if(aFindType == 12 || aFindType == 22){
                                if(l != 0 && l < iPeptide.length -1){
                                    lFoundEle2 = true;
                                    lPosition2 = l;
                                }
                            }
                            if(aFindType == 23 || aFindType == 13){
                                if( l == iPeptide.length -1){
                                    lFoundEle2 = true;
                                    lPosition2 = l;
                                }
                            }
                        }
                        if(lFoundEle2){
                            //the second element was found
                            //we found both elements
                            //check if we can find a couple in a specific mass range for an insertion
                            Vector<Couple> lReplaceCouples = getCouplesInRange(aCoupleToReplace,lMass, iMassLimit, couple);
                            for(int i = 0 ; i<lReplaceCouples.size() ; i++){
                                Couple coupleFound = lReplaceCouples.get(i);
                                double lMassFound = coupleFound.getMw();
                                if(lMass - iMassLimit < lMassFound && lMassFound < lMass + iMassLimit){
                                    //we found something in the mass interval
                                    //create a new sequence
                                    Element[] seque = iPeptide.clone();
                                    String mutOne = "";
                                    String mutTwo = "";
                                    if(aFindType == 12 && aReplaceType == 1 ){
                                        seque[lPosition1] = coupleFound.getElement1();
                                        mutOne = iPeptide[lPosition1].getElement() + " |"  + lPosition1 + "| " + coupleFound.getElement1().getElement();
                                        seque[lPosition2] = null;
                                        mutTwo = iPeptide[lPosition2].getElement() + " |"  + lPosition2 + "| .";
                                    }
                                    if(aFindType == 22 && aReplaceType == 2){
                                        seque[lPosition1] = coupleFound.getElement1();
                                        mutOne = iPeptide[lPosition1].getElement() + " |"  + lPosition1 + "| " + coupleFound.getElement1().getElement();
                                        seque[lPosition2] = null;
                                        mutTwo = iPeptide[lPosition2].getElement() + " |"  + lPosition2 + "| .";
                                    }
                                    if(aFindType == 23 && aReplaceType == 3){
                                        seque[lPosition2] = coupleFound.getElement1();
                                        mutOne = iPeptide[lPosition2].getElement() + " |"  + lPosition2 + "| "+ coupleFound.getElement1().getElement();
                                        seque[lPosition1] = null;
                                        mutTwo = iPeptide[lPosition1].getElement() + " |"  + lPosition1 + "| .";
                                    }
                                    String newSeqeunce ="";
                                    for(int s = 0; s<seque.length ; s++ ){
                                        if(seque[s] != null){
                                            newSeqeunce =  newSeqeunce + seque[s].getElement();
                                        }
                                    }

                                    newSeqVector.add(new MutatedPeptide(iPeptideSequence, newSeqeunce, "d", mutOne , mutTwo, iPeptideMass, getMass(seque),seque));

                                }
                            }
                            lFoundEle2 = false;
                        }
                    }
                    lFoundEle1 = false;
                }
            }
        }
    }


    /**
     * This method will find mutations
     * @param aCouples The couples
     * @param aType A type
     */
    public void findMutations(Vector<Couple> aCouples, int aType){
        //create a couple array
        Couple[] lCoupleArray = new Couple[aCouples.size()];
        aCouples.toArray(lCoupleArray);


        //check every couple if we can mutate it
        for(int j = 0 ; j<lCoupleArray.length ; j++){
            //get the couple
            Couple couple = lCoupleArray[j];
            Element ele1 = couple.getElement1();
            Element ele2 = couple.getElement2();
            double lMass = couple.getMw();

            //create the boolean that indicates if we found element1
            boolean lFoundEle1 = false;
            //create the boolean that indicates if we found element2
            boolean lFoundEle2 = false;
            //the position of element 1
            int lPosition1 = 0;
            //the position of element 2
            int lPosition2 = 0;

            //check every element in the peptide to mutate if we can match it on this couple
            for(int k = 0; k<iPeptide.length; k ++){
                if(ele1.getElement().equalsIgnoreCase(iPeptide[k].getElement())){
                    //check if the found position is correct with the type
                    if(aType == 12 || aType == 13){
                        if(k == 0){
                            lFoundEle1 = true;
                            lPosition1 = k;
                        }
                    }
                    //check if the found position is correct with the type
                    if(aType == 22  || aType == 23){
                        if(k != 0 && k!= iPeptide.length -1){
                            lFoundEle1 = true;
                            lPosition1 = k;
                        }
                    }
                }
                if(lFoundEle1){
                    for(int l = 0; l<iPeptide.length; l ++){
                        if(ele2.getElement().equalsIgnoreCase(iPeptide[l].getElement())){
                            //check if the found position is correct with the type
                            if(aType == 12 || aType == 22){
                                if(l != 0 && l < iPeptide.length -1){
                                    lFoundEle2 = true;
                                    lPosition2 = l;
                                }
                            }
                            //check if the found position is correct with the type
                            if(aType == 23 || aType == 13){
                                if( l == iPeptide.length -1){
                                     lFoundEle2 = true;
                                     lPosition2 = l;
                                }
                            }
                        }
                        if(lFoundEle2){
                            //we found both elements
                            //check if we can find a couple within a specific mass range
                            Vector<Couple> lReplaceCouples = getCouplesInRange(aCouples,lMass, iMassLimit,couple);
                            for(int i = 0 ; i<lReplaceCouples.size() ; i++){
                                Couple coupleFound = lReplaceCouples.get(i);
                                double lMassFound = coupleFound.getMw();
                                if(lMass - iMassLimit < lMassFound && lMassFound < lMass + iMassLimit){
                                    //we found something in the mass interval
                                    //create a new sequence
                                    Element[] seque = iPeptide.clone();
                                    //check if it's not an I to L mutation
                                    seque[lPosition1] = coupleFound.getElement1();
                                    seque[lPosition2] = coupleFound.getElement2();
                                    String newSeqeunce ="";
                                    for(int s = 0; s<seque.length ; s++ ){
                                        newSeqeunce =  newSeqeunce + seque[s].getElement();
                                    }

                                    boolean lUse = true;
                                    if(iPeptide[lPosition1].getElement().equalsIgnoreCase(coupleFound.getElement1().getElement())){
                                        lUse = false;
                                    }
                                    if(iPeptide[lPosition2].getElement().equalsIgnoreCase(coupleFound.getElement2().getElement())){
                                        lUse = false;
                                    }
                                    if(iPeptide[lPosition1].getOriginalElement().indexOf("I") > -1){
                                        if(coupleFound.getElement1().getOriginalElement().indexOf("L") > -1){
                                            lUse = false;
                                        }
                                    }
                                    if(iPeptide[lPosition1].getOriginalElement().indexOf("L") > -1){
                                        if(coupleFound.getElement1().getOriginalElement().indexOf("I") > -1){
                                            lUse = false;
                                        }
                                    }
                                    if(iPeptide[lPosition2].getOriginalElement().indexOf("I") > -1){
                                        if(coupleFound.getElement2().getOriginalElement().indexOf("L") > -1){
                                            lUse = false;
                                        }
                                    }
                                    if(iPeptide[lPosition2].getOriginalElement().indexOf("L") > -1){
                                        if(coupleFound.getElement2().getOriginalElement().indexOf("I") > -1){
                                            lUse = false;
                                        }
                                    }
                                    if(lUse){
                                        newSeqVector.add(new MutatedPeptide(iPeptideSequence, newSeqeunce, "m", iPeptide[lPosition1].getElement() + " |"  + lPosition1 + "| " + coupleFound.getElement1().getElement() , iPeptide[lPosition2].getElement() + " |"  + lPosition2 + "| " + coupleFound.getElement2().getElement(), iPeptideMass, getMass(seque),seque));
                                    }
                                }
                            }
                            lFoundEle2 = false;
                        }
                    }
                    lFoundEle1 = false;
                }
            }
        }
    }

    /**
     * Getter for the new sequences
     * @return String[] with the new sequences
     */
    public String[] getNewSequences(){
        String[] newSequences = new String[newSeqVector.size()];
        newSeqVector.toArray(newSequences);
        return newSequences;
    }

    /**
     * Getter for the new sequences
     * @return String[] with the new sequences
     */
    public Vector<MutatedPeptide> getMutatedPeptides(){
        return newSeqVector;
    }
}