package de.proteinms.xtandemparser.Mutater;
import com.compomics.peptizer.util.PeptideIdentification;
import com.compomics.peptizer.util.datatools.interfaces.PeptizerSpectrum;
import com.compomics.peptizer.util.iterators.XTandemIterator;
import de.proteinms.xtandemparser.xtandem.InputParams;
import de.proteinms.xtandemparser.xtandem.XTandemFile;
import org.xml.sax.SAXException;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Collections;
import java.util.HashMap;
import java.util.Vector;

/**
 * Created by IntelliJ IDEA.
 * User: Niklaas
 * Date: 24-Aug-2010
 * Time: 08:29:37
 * To change this template use File | Settings | File Templates.
 */
public class TandemMutater {

    private Vector<Element> iElements = new Vector<Element>();
    private Vector<String> iParameters = new Vector<String>();
    private Vector<Modification> iModifications = new Vector<Modification>();
    private Vector<Couple> iCouples = new Vector<Couple>();
    private Vector<Couple> iCouples22 = new Vector<Couple>();
    private Vector<Couple> iCouples12 = new Vector<Couple>();
    private Vector<Couple> iCouples13 = new Vector<Couple>();
    private Vector<Couple> iCouples23 = new Vector<Couple>();
    private Vector<Couple> iCouples1_ = new Vector<Couple>();
    private Vector<Couple> iCouples2_ = new Vector<Couple>();
    private Vector<Couple> iCouples3_ = new Vector<Couple>();
    private boolean iParametersEqual;

    private double iHydrogen = 1.007825;
    private double iCarbon = 12.000000;
    private double iNitrogen = 14.003074;
    private double iOxygen = 15.994915;
    private double iElectron = 0.000549;
    private double iC_term = 17.002740;
    private double iN_term = 1.007825;
    private double iThreshold = 0.05;
    private Vector<String> iModifiedSequences;
    private String iFolder;
    private int iCycles;
    private double iMassRange;
    private int iShuffleNumber;
    private File[] iTandemfiles;


    public TandemMutater(String lLocation,int lCycles, double lMassRange, int lShuffleNumbers) {
        this.iCycles = lCycles;
        this.iMassRange = lMassRange;
        this.iShuffleNumber = lShuffleNumbers;
        this.getTandemFromFolder(lLocation);
        try {
            XTandemFile iXTandemfile = new XTandemFile(iTandemfiles[0].getAbsolutePath());
            InputParams lInput = iXTandemfile.getInputParameters();
            createInitialElements();
            getModifications(lInput);
            createModifiedElements();
            createCouples();
        } catch (SAXException e) {
            e.printStackTrace();
        }

        System.gc();


    }


    public Element findElement(String lSequence) {
        for (int i = 0; i < iElements.size(); i++) {
            if (iElements.get(i).getElement().equalsIgnoreCase(lSequence)) {
                return iElements.get(i);
            }
        }
        return null;
    }

    public void getTandemFromFolder(String lFolderLocations) {
        iFolder = lFolderLocations + "\\";
        File lFolder = new File(lFolderLocations);
        File[] lFiles = lFolder.listFiles();
        Vector<File> lTempFiles = new Vector<File>();
        for (int i = 0; i < lFiles.length; i++) {
            if (lFiles[i].getAbsolutePath().endsWith(".xml")) {
                lTempFiles.add(lFiles[i]);
            }
        }
        iTandemfiles = new File[lTempFiles.size()];
        lTempFiles.toArray(iTandemfiles);
    }


    private void createCouples() {
        Vector<Element> lElements1 = new Vector<Element>();
        Vector<Element> lElements2 = new Vector<Element>();
        Vector<Element> lElements3 = new Vector<Element>();
        for (int i = 0; i < iElements.size(); i++) {
            if (iElements.get(i).getType() == 1) {
                lElements1.add(iElements.get(i));
            }
            if (iElements.get(i).getType() == 2) {
                lElements2.add(iElements.get(i));
            }
            if (iElements.get(i).getType() == 3) {
                lElements3.add(iElements.get(i));
            }
            //System.out.println(iElements.get(i));
        }

        //create 2_2 couples
        for (int j = 0; j < lElements2.size(); j++) {
            for (int k = 0; k < lElements2.size(); k++) {
                Couple lCouple = new Couple(lElements2.get(j), lElements2.get(k));
                iCouples.add(lCouple);
                iCouples22.add(lCouple);
            }
        }
        //create 1_2 couples
        for (int j = 0; j < lElements1.size(); j++) {
            for (int k = 0; k < lElements2.size(); k++) {
                Couple lCouple = new Couple(lElements1.get(j), lElements2.get(k));
                iCouples.add(lCouple);
                iCouples12.add(lCouple);
            }
        }

        //create 23 couples
        for (int j = 0; j < lElements2.size(); j++) {
            for (int k = 0; k < lElements3.size(); k++) {
                Couple lCouple = new Couple(lElements2.get(j), lElements3.get(k));
                iCouples.add(lCouple);
                iCouples23.add(lCouple);
            }
        }

        //create 13 couples
        for (int j = 0; j < lElements1.size(); j++) {
            for (int k = 0; k < lElements3.size(); k++) {
                Couple lCouple = new Couple(lElements1.get(j), lElements3.get(k));
                iCouples.add(lCouple);
                iCouples13.add(lCouple);
            }
        }

        //create 1_ couples
        for (int j = 0; j < lElements1.size(); j++) {
            Couple lCouple = new Couple(lElements1.get(j), null);
            iCouples.add(lCouple);
            iCouples1_.add(lCouple);
        }

        //create 2_ couples
        for (int j = 0; j < lElements2.size(); j++) {
            Couple lCouple = new Couple(lElements2.get(j), null);
            iCouples.add(lCouple);
            iCouples2_.add(lCouple);
        }

        //create 3_ couples
        for (int j = 0; j < lElements3.size(); j++) {
            Couple lCouple = new Couple(lElements3.get(j), null);
            iCouples.add(lCouple);
            iCouples3_.add(lCouple);
        }

        CoupleSorter lSorter = new CoupleSorter();
        Collections.sort(iCouples22, lSorter);
        Collections.sort(iCouples23, lSorter);
        Collections.sort(iCouples13, lSorter);
        Collections.sort(iCouples12, lSorter);
        Collections.sort(iCouples1_, lSorter);
        Collections.sort(iCouples2_, lSorter);
        Collections.sort(iCouples3_, lSorter);


    }


    private void getModifications(InputParams lInput){


        //get fixed mods
        String lFixMod = lInput.getResidueModMass();
        if(lFixMod.length() != 0){
            String[] lTempsMods = lFixMod.split(",");
            for(int i = 0; i<lTempsMods.length; i ++){
                double lTempMass = Double.valueOf(lTempsMods[i].substring(lTempsMods[i].indexOf("@")));
                iModifications.add(new Modification(lTempsMods[i] ,lTempMass, true));
            }
        }

        //get variable cterm mods
        String lVarMod = lInput.getResiduePotModMass();
        if(lVarMod.length() != 0){
            String[] lTempsMods = lVarMod.split(",");
            for(int i = 0; i<lTempsMods.length; i ++){
                double lTempMass = Double.valueOf(lTempsMods[i].substring(0,lTempsMods[i].indexOf("@")));
                iModifications.add(new Modification(lTempsMods[i] ,lTempMass, false));
            }
        }




        
    }

    private void createInitialElements(){
        iElements.add(new Element("A",71.037114,2));
        iElements.add(new Element("B",114.534940,2));
        iElements.add(new Element("C",103.009185,2));
        iElements.add(new Element("D",115.026943,2));
        iElements.add(new Element("E",129.042593,2));
        iElements.add(new Element("F",147.068414,2));
        iElements.add(new Element("G",57.021464,2));
        iElements.add(new Element("H",137.058912,2));
        iElements.add(new Element("I",113.084064,2));
        iElements.add(new Element("J",0.000000,2));
        iElements.add(new Element("K",128.094963,2));
        iElements.add(new Element("L",113.084064,2));
        iElements.add(new Element("M",131.040485,2));
        iElements.add(new Element("N",114.042927,2));
        iElements.add(new Element("O",0.000000,2));
        iElements.add(new Element("P",97.052764,2));
        iElements.add(new Element("Q",128.058578,2));
        iElements.add(new Element("R",156.101111,2));
        iElements.add(new Element("S",87.032028,2));
        iElements.add(new Element("T",101.047679,2));
        iElements.add(new Element("U",150.953630,2));
        iElements.add(new Element("V",99.068414,2));
        iElements.add(new Element("W",186.079313,2));
        iElements.add(new Element("X",111.000000,2));
        iElements.add(new Element("Y",163.063329,2));
        iElements.add(new Element("Z",128.550590,2));
    }
    public void createModifiedElements() {
        Vector<Element> toDelete = new Vector<Element>();


        for (int e = 0; e < iElements.size(); e++) {
            Element lEle = iElements.get(e);
            if (lEle.getType() == 2) {
                //it not modified, try to find fixed or variable internal modifications
                boolean lFixedModificationFound = false;

                for (int i = 0; i < iModifications.size(); i++) {
                    if (iModifications.get(i).isFixedModification()) {
                        Modification lMod = iModifications.get(i);
                        if (lMod.isInternal()) {
                            if (lMod.modifies(iElements.get(e).getElement())) {
                                lEle.setElement(lEle.getElement() + "<" + lMod.getConversion() + ">");
                                lEle.setMass(lEle.getMw() + lMod.getMw());
                                lEle.setInternalModified();
                                lFixedModificationFound = true;
                            }
                        }
                    }
                }

                if (!lFixedModificationFound) {
                    //no fixed modification found find variable modifications
                    for (int i = 0; i < iModifications.size(); i++) {
                        Modification lMod = iModifications.get(i);
                        if (lMod.isInternal()) {
                            //only to the internal modifications for now
                            if (lMod.modifies(iElements.get(e).getElement())) {
                                //it cannot be already internally modified
                                String lName = lEle.getElement();
                                double lMass = lEle.getMw();
                                int lType = lEle.getType();
                                if (lType != 3) {
                                    lName = lName + "<" + lMod.getConversion() + ">";
                                } else {
                                    lName = lName.replace("-", "<" + lMod.getConversion() + ">-");
                                }
                                lMass = lMass + lMod.getMw();
                                Element lModifiedElement = new Element(lName, lMass, lType, lEle.getOriginalElement());
                                lModifiedElement.setInternalModified();
                                iElements.add(lModifiedElement);
                            }
                        }
                    }
                }


                //add terminal modified elements that will be deleted


                for (int i = 0; i < iModifications.size(); i++) {
                    Modification lMod = iModifications.get(i);
                    if (lMod.isNterm()) {
                        if (lMod.getAminoAcids().size() > 0) {
                            if (lMod.modifies(lEle.getOriginalElement())) {
                                if (!lEle.isModified()) {
                                    String lName = lEle.getElement() + "<" + lMod.getConversion() + ">";
                                    double lMass = lEle.getMw() + lMod.getMw();
                                    Element lModifiedElement = new Element(lName, lMass, 2, lEle.getOriginalElement());
                                    lModifiedElement.setInternalModified();
                                    lModifiedElement.setNtermRestriction();
                                    iElements.add(lModifiedElement);
                                    toDelete.add(lModifiedElement);
                                } else {
                                    //check if the it is not the same modification
                                    if(!lEle.getElement().equalsIgnoreCase(lEle.getOriginalElement() + "<" + lMod.getConversion() + ">")){
                                        if(!lEle.isDoubleModified()){
                                            String lName = lEle.getElement().replace("<", "<" + lMod.getConversion() + ",");
                                            double lMass = lEle.getMw() + lMod.getMw();
                                            Element lModifiedElement = new Element(lName, lMass, 2, lEle.getOriginalElement());
                                            lModifiedElement.setInternalModified();
                                            lModifiedElement.setNtermRestriction();
                                            lModifiedElement.setDoubleModified();
                                            iElements.add(lModifiedElement);
                                            toDelete.add(lModifiedElement);
                                        }
                                    }
                                }
                            }
                        }
                    }
                    if (lMod.isCterm()) {
                        if (lMod.getAminoAcids().size() > 0) {
                            if (lMod.modifies(lEle.getOriginalElement())) {
                                if (!lEle.isModified()) {
                                    String lName = lEle.getElement() + "<" + lMod.getConversion() + ">";
                                    double lMass = lEle.getMw() + lMod.getMw();
                                    Element lModifiedElement = new Element(lName, lMass, 2, lEle.getOriginalElement());
                                    lModifiedElement.setInternalModified();
                                    lModifiedElement.setCtermRestriction();
                                    iElements.add(lModifiedElement);
                                    toDelete.add(lModifiedElement);
                                } else {
                                    String lName = lEle.getElement().replace(">", "," + lMod.getConversion() + ">");
                                    double lMass = lEle.getMw() + lMod.getMw();
                                    Element lModifiedElement = new Element(lName, lMass, 2, lEle.getOriginalElement());
                                    lModifiedElement.setInternalModified();
                                    lModifiedElement.setNtermRestriction();
                                    iElements.add(lModifiedElement);
                                    toDelete.add(lModifiedElement);
                                }
                            }
                        }
                    }
                }

                boolean lFixedCtermModification = false;
                Modification lFixedCtermMod = null;
                boolean lFixedNtermModification = false;
                Modification lFixedNtermMod = null;
                for (int i = 0; i < iModifications.size(); i++) {
                    if (iModifications.get(i).isFixedModification()) {
                        Modification lMod = iModifications.get(i);
                        if(lMod.getAminoAcids().size() != 0){
                             if (lMod.modifies(lEle.getOriginalElement())) {
                                if (lMod.isFixedModification()) {
                                    if (lMod.isNterm()) {
                                        lFixedNtermModification = true;
                                        lFixedNtermMod = lMod;
                                    }
                                    if (lMod.isCterm()) {
                                        lFixedCtermModification = true;
                                        lFixedCtermMod = lMod;
                                    }
                                }
                            }
                        }
                    }
                }
                if (!lEle.isCtermRestriction()) {

                    if (lFixedNtermModification) {
                        //add the fixed modification
                        String lName = lFixedNtermMod.getConversion() + "-" + lEle.getElement();
                        double lMass = lEle.getMw() + lFixedNtermMod.getMw();
                        Element lModifiedElement = new Element(lName, lMass, 1, lEle.getOriginalElement());
                        lModifiedElement.setNtermModified();
                        iElements.add(lModifiedElement);
                    } else {
                        //no fixed modifications
                        //add the variable modifications
                        //add the NH2- "modification"
                        if(!lEle.isNtermRestriction()){
                            for (int i = 0; i < iModifications.size(); i++) {
                                if (!iModifications.get(i).isFixedModification()) {
                                    Modification lMod = iModifications.get(i);
                                    if (!lMod.isFixedModification() && lMod.isNterm()) {
                                        if (lMod.getAminoAcids().size() == 0) {
                                            String lName = lMod.getConversion() + "-" + lEle.getElement();
                                            double lMass = lEle.getMw() + lMod.getMw();
                                            Element lModifiedElement = new Element(lName, lMass, 1, lEle.getOriginalElement());
                                            lModifiedElement.setNtermModified();
                                            iElements.add(lModifiedElement);
                                        }
                                    }
                                }
                            }
                        }
                        String lName = "NH2-" + lEle.getElement();
                        double lMass = lEle.getMw() + iN_term;
                        Element lModifiedElement = new Element(lName, lMass, 1, lEle.getOriginalElement());
                        lModifiedElement.setNtermModified();
                        iElements.add(lModifiedElement);
                    }
                }

                if (!lEle.isNtermRestriction()) {

                    if (lFixedCtermModification) {
                        //add the fixed modification
                        String lName = lEle.getElement() + "-" + lFixedCtermMod.getConversion();
                        double lMass = lEle.getMw() + lFixedCtermMod.getMw();
                        Element lModifiedElement = new Element(lName, lMass, 3, lEle.getOriginalElement());
                        lModifiedElement.setNtermModified();
                        iElements.add(lModifiedElement);
                    } else {
                        //no fixed modifications
                        //add the variable modifications
                        //add the -COOH "modification"
                        if(!lEle.isCtermRestriction()){
                            for (int i = 0; i < iModifications.size(); i++) {
                                if (iModifications.get(i).isFixedModification()) {
                                    Modification lMod = iModifications.get(i);
                                    if (!lMod.isFixedModification() && lMod.isCterm()) {
                                        String lName = lEle.getElement() + "-" + lMod.getConversion();
                                        double lMass = lEle.getMw() + lMod.getMw();
                                        Element lModifiedElement = new Element(lName, lMass, 3, lEle.getOriginalElement());
                                        lModifiedElement.setNtermModified();
                                        iElements.add(lModifiedElement);
                                    }
                                }
                            }
                        }
                        String lName = lEle.getElement() + "-COOH";
                        double lMass = lEle.getMw() + iC_term;
                        Element lModifiedElement = new Element(lName, lMass, 3, lEle.getOriginalElement());
                        lModifiedElement.setNtermModified();
                        iElements.add(lModifiedElement);
                    }
                }

            }



        }

        for (int i = 0; i < toDelete.size(); i++) {
            iElements.remove(toDelete.get(i));
        }
    }


    private void getIdentificationsAndSpectra() throws IOException, SAXException {
        Vector<String> lModSeq = new Vector<String>();
        for (int f = 0; f < iTandemfiles.length; f++) {
            //create new spectrum writer
            BufferedWriter out = new BufferedWriter(new FileWriter(iTandemfiles[f].getParentFile() + "\\" + f + "merge.mgf"));

            int lCounter = 0;
            int lCounterNot = 0;
            int lCounterMod = 0;

            XTandemIterator lIt = new XTandemIterator(iTandemfiles[f]);
            while(lIt.hasNext()){
                PeptideIdentification lId = lIt.next();

                if (lId.getNumberOfConfidentPeptideHits() >= 1) {
                    boolean lNew = true;

                    String lValue = lId.getBestPeptideHit().getModifiedSequence();
                    for (int m = 0; m < lModSeq.size(); m++) {
                        if (lValue.equalsIgnoreCase(lModSeq.get(m))) {
                            lNew = false;
                            m = lModSeq.size();
                        }
                    }
                    if (lNew) {
                        lModSeq.add(lValue);
                    }

                    PeptizerSpectrum lSpectrumPeptizer = lId.getSpectrum();
                    String lFileName = lSpectrumPeptizer.getName();
                    lCounter = lCounter + 1;
                    String lSpectrumAsString = "\nBEGIN IONS\nTITLE=" + lSpectrumPeptizer.getName() + "\n";
                    lSpectrumAsString = lSpectrumAsString + "PEPMASS=" + lSpectrumPeptizer.getPrecursorMZ() + "\n";
                    lSpectrumAsString = lSpectrumAsString + "CHARGE=" + lSpectrumPeptizer.getChargeString() + "\n";
                    for (int i = 0; i < lSpectrumPeptizer.getPeakList().length; i++) {
                        lSpectrumAsString = lSpectrumAsString + lSpectrumPeptizer.getPeakList()[i].getMZ() + "\t" + lSpectrumPeptizer.getPeakList()[i].getIntensity() + "\n";
                    }
                    lSpectrumAsString = lSpectrumAsString + "END IONS\n";
                    out.write(lSpectrumAsString);

                    if (lCounter % 50 == 0) {
                        out.flush();
                        System.out.print(".");
                    }
                    if (lCounter % 2000 == 0) {
                        out.flush();
                        System.out.print(".\n");
                    }

                } else {
                    lCounterNot = lCounterNot + 1;
                }
                //System.out.println(lId);

            }

            out.flush();
            out.close();
            System.gc();
        }
        iModifiedSequences = lModSeq;
    }

    public void doMutations() throws IOException {
        FastaDbSaver lFastaDbSaver = new FastaDbSaver(iFolder + "result.fasta");
        //create new summary writer
        BufferedWriter iOut = new BufferedWriter(new FileWriter(iFolder + "summary.txt"));

        //start mutating
        try {
            iOut.write("# mutations found , sequence , sequence length, different elements\n");

            int lTotalCount = 0;
            int lLastCount = 0;
            for (int i = 0; i < iModifiedSequences.size(); i++) {
                HashMap lPrevious = new HashMap();

                lTotalCount = doPeptideMutation(lTotalCount, lPrevious, iModifiedSequences.get(i), iCycles, iMassRange, lFastaDbSaver, iShuffleNumber, 0);


                iOut.write((lTotalCount - lLastCount) + "," + iModifiedSequences.get(i) + "\n");
                iOut.flush();
                lLastCount = lTotalCount;

            }
            lFastaDbSaver.closeWriter();
            iOut.write("Mutated " + iModifiedSequences.size() + " peptides! : " + lTotalCount + "\n");
            iOut.flush();
            iOut.close();

        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    /**
     * This method will create an array of elements for a modified peptide sequence
     *
     * @param sequence Ths sequence to split
     * @return Element[]
     */
    public Element[] sequenceToArray(String sequence) {
        //create a vector for the different sequences
        Vector<Element> sequenceVector = new Vector<Element>();

        String sub1 = "";
        if (sequence.indexOf(('<'), sequence.indexOf('-')) - sequence.indexOf('-') == 2) {
            sequenceVector.add(findElement(sequence.substring(0, sequence.indexOf(">") + 1)));
            sub1 = sequence.substring(sequence.indexOf(">") + 1);
        } else {
            sequenceVector.add(findElement(sequence.substring(0, sequence.indexOf("-") + 2)));
            sub1 = sequence.substring(sequence.indexOf("-") + 2);
        }

        for (int i = 0; i < sub1.length(); i++) {
            if (sub1.charAt(i + 1) == '<') {
                if (sub1.charAt(sub1.indexOf(">", i) + 1) == '-') {
                    sequenceVector.add(findElement(sub1.substring(i)));
                    i = sub1.length() - 1;
                } else {
                    sequenceVector.add(findElement(sub1.substring(i, sub1.indexOf(">", i) + 1)));
                    i = sub1.indexOf(">", i);
                }


            } else {
                if (sub1.charAt(i + 1) == '-') {
                    sequenceVector.add(findElement(sub1.substring(i)));
                    i = sub1.length() - 1;
                } else {
                    sequenceVector.add(findElement(String.valueOf(sub1.charAt(i))));
                }
            }
        }
        Element[] sequenceArray = new Element[sequenceVector.size()];
        sequenceVector.toArray(sequenceArray);
        return sequenceArray;
    }

    /**
     * The method will find the non redundant sequences in the mutation result array. It will also check
     * if the mutation result is already stored. It will also recheck the masses.
     *
     * @param lMutationResults          The result of the mutations
     * @param lPreviouslyStoredPeptides The previously stored peptides
     * @param lMassRange                The mass range
     * @return String[] with the non redundant peptides
     */
    public Vector<MutatedPeptide> clearRedundancyOfPeptides(Vector<MutatedPeptide> lMutationResults, HashMap lPreviouslyStoredPeptides, double lMassRange) {
        //create a vector to store all non redundant sequences in
        Vector<MutatedPeptide> lNonRedundantSeqVector = new Vector<MutatedPeptide>();

        //check all the results of the mutation
        for (int i = 0; i < lMutationResults.size(); i++) {
            //get the mutation result
            String lMutation = lMutationResults.get(i).getMutatedSequence();
            //check if it is stored in a previous mutation cycle
            boolean found = lPreviouslyStoredPeptides.containsKey(lMutation);
            if (!found) {
                //we did not find it
                //get the mass for an extra mass check
                Element[] lMutationPeptideObject = sequenceToArray(lMutation);
                double lMutationMass = lMutationResults.get(i).getMutatedMass();
                double lOriginalMass = lMutationResults.get(i).getOriginalMass();
                if (lMutationMass > lOriginalMass - lMassRange && lMutationMass < lOriginalMass + lMassRange) {
                    lNonRedundantSeqVector.add(lMutationResults.get(i));
                    lPreviouslyStoredPeptides.put(lMutation, 1);
                }
            }
        }
        return lNonRedundantSeqVector;
    }

    public int doPeptideMutation(int lCounter, HashMap lPreviouslyStoredPeptides, String lMutationPeptideString, int lCycles, double lMassRange, FastaDbSaver lFastaDbSaver, int lShuffleNumber, int lInternalCounter) {
        //create an Object[] for the peptide that must be mutated
        Element[] lMutationPeptide = sequenceToArray(lMutationPeptideString);
        for (int i = 0; i < lMutationPeptide.length; i++) {
            if (lMutationPeptide[i] == null) {
                System.out.println("Error");
            }
        }
        //create the mutation finder
        MutationFinder lMutationFinder = new MutationFinder(lMutationPeptide, lMassRange);
        //find mutations
        lMutationFinder.findMutations(iCouples12, 12);
        lMutationFinder.findMutations(iCouples22, 22);
        lMutationFinder.findMutations(iCouples23, 23);
        lMutationFinder.findMutations(iCouples13, 13);
        //find deletions
        lMutationFinder.findDeletion(iCouples12, iCouples1_, 12, 1);
        lMutationFinder.findDeletion(iCouples22, iCouples2_, 22, 2);
        lMutationFinder.findDeletion(iCouples23, iCouples3_, 23, 3);
        //find insertions
        lMutationFinder.findInsertion(iCouples1_, iCouples12, 1, 12);
        lMutationFinder.findInsertion(iCouples2_, iCouples22, 2, 22);
        lMutationFinder.findInsertion(iCouples3_, iCouples23, 3, 23);

        //get the result
        Vector<MutatedPeptide> lMutatedPeptides = lMutationFinder.getMutatedPeptides();
        //clear redundancy
        Vector<MutatedPeptide> nonRedundantPeptides = clearRedundancyOfPeptides(lMutatedPeptides, lPreviouslyStoredPeptides, lMassRange);

        for (int j = 0; j < nonRedundantPeptides.size(); j++) {
            MutatedPeptide lMutatedPeptide = nonRedundantPeptides.get(j);
            Element[] nonRedundantMutatedPeptide = lMutatedPeptide.getMutationResult();
            //final check if it is not the same
            if (!lMutatedPeptide.getMutatedSequence().equalsIgnoreCase(lMutationPeptideString)) {
                //shuffle and write the shuffled mutated peptides
                for (int k = 0; k < lShuffleNumber; k++) {

                    //check if the shuffled peptides is not in the previously stored hashmap
                    Element[] lShuffle = shuffle(nonRedundantMutatedPeptide);
                    int lShuffleCounter = 0;
                    while (lPreviouslyStoredPeptides.containsKey(lShuffle) && lShuffleCounter < 100) {
                        lShuffle = shuffle(nonRedundantMutatedPeptide);
                        lShuffleCounter = lShuffleCounter + 1;
                    }

                    if (lShuffleCounter < 100) {
                        //add the shuffled peptide to the previously stored hashmap
                        String lHeader = ">" + lCounter;
                        lHeader = lHeader + "^c" + lCycles + "." + lInternalCounter + "s" + lShuffleCounter + "^" + lMutatedPeptide.getMutationType() + " " + lMutatedPeptide.getMutationOne() + " " + lMutatedPeptide.getMutationTwo();
                        lFastaDbSaver.writeMutatedePeptidesInDb(lHeader);
                        lFastaDbSaver.writeMutatedePeptidesInDb(removeModifications(lShuffle));
                        lFastaDbSaver.writeMutatedePeptidesInDb("");
                        lPreviouslyStoredPeptides.put(lShuffle, 1);
                        lCounter = lCounter + 1;
                        lInternalCounter = lInternalCounter + 1;
                    } else {
                        // we cannot find any shuffled stuff
                        System.out.println("Unable to shuffle the peptide '" + lMutationPeptideString + "' more then " + k + " times");
                        //so, stop the for loop
                        k = lShuffleNumber;
                    }

                }

                //write the mutated peptide
                String lHeader = ">" + lCounter;
                lHeader = lHeader + "^c" + lCycles + "." + lInternalCounter + "^" + lMutatedPeptide.getMutationType() + " " + lMutatedPeptide.getMutationOne() + " " + lMutatedPeptide.getMutationTwo();
                lFastaDbSaver.writeMutatedePeptidesInDb(lHeader);
                lFastaDbSaver.writeMutatedePeptidesInDb(removeModifications(lMutatedPeptide.getMutationResult()));
                lFastaDbSaver.writeMutatedePeptidesInDb("");
                lCounter = lCounter + 1;
                lInternalCounter = lInternalCounter + 1;
            }
        }

        //check if we have to do an extra mutation cycle
        lCycles = lCycles - 1;
        if (lCycles > 0) {
            //an extra mutation cycle is needed
            for (int k = 0; k < nonRedundantPeptides.size(); k++) {
                int lTempCounter = lCounter;
                lCounter = doPeptideMutation(lTempCounter, lPreviouslyStoredPeptides, lMutatedPeptides.get(k).getMutatedSequence(), lCycles, lMassRange, lFastaDbSaver, lShuffleNumber, lInternalCounter);
            }
        }
        return lCounter;
    }


    public String removeModifications(Element[] peptide) {
        String lRemoved = "";
        for (int i = 0; i < peptide.length; i++) {
            if (peptide[i] != null) {
                lRemoved = lRemoved + peptide[i].getOriginalElement();
            }
        }
        return lRemoved;
    }


    /**
     * This method will shuffle the given object array (not the first and the last elements)
     *
     * @param sequence Object[] to shuffle
     * @return String with the shuffled sequence
     */
    public Element[] shuffle(Element[] sequence) {
        Element[] seqToShuffle = new Element[sequence.length];
        int replaceLocation = 0;
        boolean[] replaceLocationUsed = new boolean[sequence.length - 2];
        for (int n = 0; n < replaceLocationUsed.length; n++) {
            replaceLocationUsed[n] = false;
        }

        for (int i = 1; i < sequence.length - 1; i++) {
            boolean ok = false;
            while (!ok) {
                replaceLocation = (int) (Math.random() * (sequence.length - 2));
                if (!replaceLocationUsed[replaceLocation]) {
                    replaceLocationUsed[replaceLocation] = true;
                    ok = true;
                }
            }

            seqToShuffle[replaceLocation + 1] = sequence[i];

        }
        return seqToShuffle;

    }




    public static void main(String[] args){
        try {
            String lLocation = "C:\\niklaas\\data\\08_10\\tandem\\CPTAC\\result\\fw\\test\\";
            String[] lLocations = new String[]{lLocation};
            TandemMutater lMut = new TandemMutater(lLocation, 1, 0.1, 0);
            lMut.getIdentificationsAndSpectra();
            lMut.doMutations();
        } catch (IOException e) {
            e.printStackTrace();
        } catch (SAXException e) {
            e.printStackTrace();
        }

    }
}
